import itertools
from collections.abc import Generator, Iterator
from dataclasses import dataclass
from typing import Literal

import anndata as ad
import narwhals as nw
import pandas as pd
from more_itertools import first, nth
from narwhals.group_by import GroupBy as NwGroupBy
from narwhals.typing import IntoDataFrameT

from annsel.core.typing import Predicates
from annsel.core.utils import _construct_adata_from_indices

# Define the named tuple at module level


@dataclass(frozen=True, slots=True)
class GroupResult:
    key: tuple[str, ...]
    indices: pd.Index

    def __iter__(self):
        yield from (self.key, self.indices)


@nw.narwhalify
def _group_by_df(df: IntoDataFrameT, *predicates: Predicates) -> NwGroupBy:
    return df.group_by(*predicates)


def _group_by_obs(adata: ad.AnnData, *predicates: Predicates) -> Generator[GroupResult, None, None]:
    return (
        GroupResult(first(gb), nth(gb, 1).to_native().index)
        for gb in _group_by_df(adata.obs, *(p.names for p in predicates))
    )


def _group_by_var(adata: ad.AnnData, *predicates: Predicates) -> Generator[GroupResult, None, None]:
    return (
        GroupResult(first(gb), nth(gb, 1).to_native().index)
        for gb in _group_by_df(adata.var, *(p.names for p in predicates))
    )


def _group_by(
    adata: ad.AnnData,
    obs: Predicates | None = None,
    var: Predicates | None = None,
    return_group_names: bool = False,
    sparse_method: Literal["csr", "csc", True, False] | None = None,
) -> Iterator[ad.AnnData | tuple]:
    # Create single-item iterables for None cases
    obs_gb = _group_by_obs(adata, obs) if obs else [(None, adata.obs_names)]
    var_gb = _group_by_var(adata, var) if var else [(None, adata.var_names)]

    # Convert to lists to ensure they can be iterated multiple times
    obs_gb = list(obs_gb)
    var_gb = list(var_gb)

    # If either grouping is empty, yield nothing
    if not obs_gb or not var_gb:
        return

    for (obs_groups, obs_idx), (var_groups, var_idx) in itertools.product(obs_gb, var_gb):
        _adata = _construct_adata_from_indices(adata, obs_idx=obs_idx, var_idx=var_idx, sparse_method=sparse_method)

        if not return_group_names:
            yield _adata
        else:
            match (obs, var):
                case (None, None):
                    yield _adata
                case (None, _):
                    yield (var_groups, _adata)
                case (_, None):
                    yield (obs_groups, _adata)
                case (_, _):
                    yield (obs_groups, var_groups, _adata)



from importlib.metadata import version

from . import datasets
from .core import col, obs_names, register_anndata_accessor, var_names
from .tl import AnnselAccessor

__all__ = [
    "AnnselAccessor",
    "col",
    "datasets",
    "obs_names",
    "register_anndata_accessor",
    "var_names",
]

__version__ = version("annsel")



from collections.abc import Callable, Iterable
from typing import Any, Protocol

import narwhals as nw
from narwhals.utils import flatten


class AnnselExpr(nw.Expr):
    """A wrapper for the `narwhals.Expr` class."""

    def __init__(self, call: Callable[[Any], Any], *names: Iterable[str]) -> None:
        super().__init__(call)
        self.names = names


class ColProtocol(Protocol):
    """Protocol for column selection."""

    def __call__(self, *names: str | Iterable[str]) -> AnnselExpr:
        """Protocol for column selection.

        Parameters
        ----------
        *names
            Names of columns to select. Can be strings or iterables of strings.

        Returns
        -------
        AnnselExpr
            A wrapper around the narwhals.Expr class for column selection.
        """
        ...


class Col:
    """A wrapper around the `narwhals.col` function."""

    names: Iterable[str]

    def __call__(self, *names: str | Iterable[str]) -> AnnselExpr:
        """
        Select columns from a DataFrame component of an AnnData object.

        Parameters
        ----------
        *names
            Names of columns to select. Can be strings or iterables of strings.

        Returns
        -------
        A wrapper around the `narwhals.Expr` class.
        """

        def column_selector(plx: Any):
            return plx.col(*flatten(names))

        return AnnselExpr(column_selector, *flatten(names))


col: Col = Col()
obs_names: AnnselExpr = col("obs_names")
var_names: AnnselExpr = col("var_names")



import warnings

import anndata as ad


class AccessorRegistrationWarning(UserWarning):
    """Warning issued when a conflicting accessor is registered."""


class _CachedAccessor:
    """Custom property-like object (descriptor) for caching accessors."""

    def __init__(self, name, accessor):
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            # we're accessing the attribute of the class, i.e., Dataset.geo
            return self._accessor

        # Use the same dict as @pandas.util.cache_readonly.
        # It must be explicitly declared in obj.__slots__.
        try:
            cache = obj._cache
        except AttributeError:
            cache = obj._cache = {}

        try:
            return cache[self._name]
        except KeyError:
            pass

        try:
            accessor_obj = self._accessor(obj)
        except AttributeError as err:
            # __getattr__ on data object will swallow any AttributeErrors
            # raised when initializing the accessor, so we need to raise as
            # something else (GH933):
            raise RuntimeError(f"error initializing {self._name!r} accessor.") from err

        cache[self._name] = accessor_obj
        return accessor_obj


def _register_accessor(name, cls):
    def decorator(accessor):
        if hasattr(cls, name):
            warnings.warn(
                f"registration of accessor {accessor!r} under name {name!r} for type {cls!r} is "
                "overriding a preexisting attribute with the same name.",
                AccessorRegistrationWarning,
                stacklevel=2,
            )
        setattr(cls, name, _CachedAccessor(name, accessor))
        return accessor

    return decorator


def register_anndata_accessor(name: str):
    """Register a custom accessor on :class:`~anndata.AnnData` objects.

    Parameters
    ----------
    name
        Name under which the accessor should be registered. A warning is issued
        if this name conflicts with a preexisting attribute.

    See Also
    --------
    xarray.register_dataset_accessor
    xarray.register_dataarray_accessor
    """
    return _register_accessor(name, ad.AnnData)



#!/usr/bin/env python3

import argparse
import subprocess
from pathlib import Path


def export_html_wasm(notebook_path: str, output_dir: str, as_app: bool = False) -> bool:
    """Export a single marimo notebook to HTML format.

    Returns
    -------
        bool
            True if export succeeded, False otherwise
    """
    output_path = notebook_path.replace(".py", ".html")

    cmd = ["marimo", "export", "html-wasm"]
    if as_app:
        print(f"Exporting {notebook_path} to {output_path} as app")
        cmd.extend(["--mode", "run", "--no-show-code"])
    else:
        print(f"Exporting {notebook_path} to {output_path} as notebook")
        cmd.extend(["--mode", "edit"])

    try:
        output_file = str(Path(output_dir) / output_path)
        Path(output_file).parent.mkdir(parents=True, exist_ok=True)

        cmd.extend([notebook_path, "-o", output_file])
        subprocess.run(cmd, capture_output=True, text=True, check=True)
        return True
    except subprocess.CalledProcessError as e:
        print(f"Error exporting {notebook_path}:")
        print(e.stderr)
        return False
    except Exception as e:  # noqa: BLE001
        print(f"Unexpected error exporting {notebook_path}: {e}")
        return False


def generate_index(all_notebooks: list[str], output_dir: str) -> None:
    """Generate the index.html file."""
    print("Generating index.html")

    index_path = Path(output_dir) / "index.html"
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    try:
        with Path.open(index_path, "w") as f:
            f.write(
                """<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />
    <title>marimo</title>
    <link href="https://cdn.jsdelivr.net/npm/tailwindcss@2.2.19/dist/tailwind.min.css" rel="stylesheet">
  </head>
  <body class="font-sans max-w-2xl mx-auto p-8 leading-relaxed">
    <div class="mb-8">
      <img src="https://raw.githubusercontent.com/marimo-team/marimo/main/docs/_static/marimo-logotype-thick.svg" alt="marimo" class="h-20" />
    </div>
    <div class="grid gap-4">
"""
            )
            for notebook in all_notebooks:
                notebook_name = notebook.split("/")[-1].replace(".py", "")
                display_name = notebook_name.replace("_", " ").title()

                f.write(
                    f'      <div class="p-4 border border-gray-200 rounded">\n'
                    f'        <h3 class="text-lg font-semibold mb-2">{display_name}</h3>\n'
                    f'        <div class="flex gap-2">\n'
                    f'          <a href="{notebook.replace(".py", ".html")}" class="px-3 py-1 bg-gray-100 hover:bg-gray-200 rounded">Open Notebook</a>\n'
                    f"        </div>\n"
                    f"      </div>\n"
                )
            f.write(
                """    </div>
  </body>
</html>"""
            )
    except OSError as e:
        print(f"Error generating index.html: {e}")


def main() -> None:
    """Build marimo notebooks."""
    parser = argparse.ArgumentParser(description="Build marimo notebooks")
    parser.add_argument("--output-dir", default="_site", help="Output directory for built files")
    args = parser.parse_args()

    all_notebooks: list[str] = []
    for directory in ["marimo/notebooks"]:
        dir_path = Path(directory)
        if not dir_path.exists():
            print(f"Warning: Directory not found: {dir_path}")
            continue

        all_notebooks.extend(str(path) for path in dir_path.rglob("*.py"))

    if not all_notebooks:
        print("No notebooks found!")
        return

    # Export notebooks sequentially
    for nb in all_notebooks:
        export_html_wasm(nb, args.output_dir, as_app=False)

    # Generate index only if all exports succeeded
    generate_index(all_notebooks, args.output_dir)


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "narwhals==1.22",
#     "marimo",
#     "anndata==0.11.3",
#     "annsel==0.0.4",
# ]
# ///
import marimo

__generated_with = "0.10.13"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# All of Annsel""")
    return


@app.cell
def _():
    import marimo as mo

    return (mo,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        This notebook tells you just about everything you need to use `annsel`. It's a good starting point to get a feel for the package.

        :::{note}
        :class: dropdown

        You should be familiar with [`AnnData` ](https://anndata.readthedocs.io/en/latest/) beforehand.
        :::
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Set up Data""")
    return


@app.cell
def _():
    # '%load_ext autoreload\n%autoreload 2' command supported automatically in marimo
    return


@app.cell
def _():
    import annsel as an

    return (an,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""We will load a Leukemic bone marrow cytometry dataset :cite:p:`Triana2021` You can view the dataset on [CellxGene](https://cellxgene.cziscience.com/e/b3a5a10f-b1cb-4e8e-abce-bf345448625b.cxg/)."""
    )
    return


@app.cell
def _(an):
    adata = an.datasets.leukemic_bone_marrow_dataset()
    return (adata,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""View the contents of the `AnnData` object.""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        Importing `annsel` will automatically register the accessor and add the `an` attribute to the `AnnData` object's namespace.

        You can access the methods of the accessor using the `an` attribute on an `AnnData` object.

        ```python

        anndata.AnnData.an.<method_name>
        ```
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Filter


        We can filter the `AnnData` object using the `filter` method. Filtering can be applied to the `obs`, `var`, `X` (with a layer), `obs_names` and `var_names`.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Let's filter the `AnnData` object by the Cell Type stored in `obs["Cell_label"]`.""")
    return


@app.cell
def _(adata, an):
    adata.an.filter(obs=an.col(["Cell_label"]) == "CD8+CD103+ tissue resident memory T cells")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        :::{note}
        This is equivalent to

        ```python
        adata[adata.obs["Cell_label"] == "CD8+CD103+ tissue resident memory T cells", :]
        ```
        :::
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""You can also apply other expressions to `an.col` callable, for example `is_in` allows you to filter by a list of values. Under the hood, `annsel` uses [Narwhals](https://narwhals-dev.github.io/narwhals/) to apply these expressions. A full list of expressions which can be applied to a column can be found [here](https://narwhals-dev.github.io/narwhals/api-reference/expr/)."""
    )
    return


@app.cell
def _(adata, an):
    adata.an.filter(
        obs=an.col(["Cell_label"]).is_in(["Classical Monocytes", "CD8+CD103+ tissue resident memory T cells"])
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""We can also combine multiple Predicates using the `&` and `|` operators.""")
    return


@app.cell
def _(adata, an):
    adata.an.filter(
        obs=(an.col(["Cell_label"]).is_in(["Classical Monocytes", "CD8+CD103+ tissue resident memory T cells"]))
        & (an.col(["sex"]) == "male")
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Or if you pass a tuple of Predicates, it will apply the `&` operator between them automatically.""")
    return


@app.cell
def _(adata, an):
    adata.an.filter(
        obs=(
            an.col(["Cell_label"]).is_in(["Classical Monocytes", "CD8+CD103+ tissue resident memory T cells"]),
            an.col(["sex"]) == "male",
        )
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""The `|` operator is applied between the Predicates as well. Here we will select all the cells that are either Classical Monocytes or CD8+CD103+ tissue resident memory T cells, or are from male samples irrespective of the cell type."""
    )
    return


@app.cell
def _(adata, an):
    adata.an.filter(
        obs=(
            an.col(["Cell_label"]).is_in(["Classical Monocytes", "CD8+CD103+ tissue resident memory T cells"])
            | (an.col(["sex"]) == "male")
        )
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""We can also filter the `AnnData` object by the `var` column. Here we will filter the `AnnData` object to only include the genes with `vst.mean` greater than or equal to 3."""
    )
    return


@app.cell
def _(adata, an):
    adata.an.filter(var=an.col(["vst.mean"]) >= 3)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""Filtering can also be applied to the `X` matrix. Here we will filter the `AnnData` object to only include the cells with `ENSG00000205336` gene expression greater than 1. If you want to filter by a layer, you can pass the layer name to the `layer` argument. and the operation will be applied to the `var_name` of that layer."""
    )
    return


@app.cell
def _(adata, an):
    adata.an.filter(
        x=an.col(["ENSG00000205336"]) > 1,
        layer=None,
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""This can all be combined together as well""")
    return


@app.cell
def _(adata, an):
    adata.an.filter(
        obs=(
            an.col(["Cell_label"]).is_in(["Classical Monocytes", "CD8+CD103+ tissue resident memory T cells"]),
            an.col(["sex"]) == "male",
        ),
        var=an.col(["vst.mean"]) >= 3,
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""Filtering can also be applied to `var_names` and `obs_names` using the `an.var_names` and `an.obs_names` predicates respectively. These are special predicates which can only be applied to `var_names` and `obs_names` of the `AnnData` object."""
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""Here, arbitrarily, we will filter the AnnData object to only include the cells with `obs_names` starting with `645` and the genes with `var_names` starting with `ENSG0000018`."""
    )
    return


@app.cell
def _(adata, an):
    adata.an.filter(
        obs_names=an.obs_names.str.starts_with("645"), var_names=an.var_names.str.starts_with("ENSG0000018")
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Select


        We can also apply the `select` method to the `AnnData` object. This is similar to the `filter` method, but it will only keep the rows and columns that match the Predicates. It can be applied to the `obs`, `var`, and `X`.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""Here we will select the `Cell_label` and `sex` columns from the `obs` table, the `feature_name` column from the `var` table and the `ENSG00000205336` gene from `X`. This will return a new `AnnData` object with only these columns in the `obs`, `var` and `X` tables."""
    )
    return


@app.cell
def _(adata, an):
    adata.an.select(
        obs=an.col(["Cell_label", "sex"]),
        var=an.col(["feature_name"]),
        x=an.col(["ENSG00000205336"]),
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Group By


        We can also group the `AnnData` object by the `obs` and `var` columns. This will return a generator of `AnnData` objects subset on each group.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""Here we will group the `AnnData` object by the `Cell_label` column in the `obs` table and the `feature_type` column in the `var` table."""
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""If you pass `return_group_names=True`, the generator will yield a tuple of the group name and the `AnnData` object. If you group by both `obs` and `var`, the generator will yield a tuple of the both group names and the `AnnData` object, if you group by only one, it will yield a tuple of the group name and the `AnnData` object."""
    )
    return


@app.cell
def _(adata, an):
    for i in adata.an.group_by(
        obs=an.col(["Cell_label"]),
        var=an.col(["feature_type"]),
        return_group_names=True,
    ):
        obs_group, var_group, _adata = i
    return i, obs_group, var_group


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()



from typing import Literal

import anndata as ad
import narwhals as nw
import pandas as pd
from narwhals.typing import IntoDataFrameT

from annsel.core.typing import Predicates
from annsel.core.utils import _construct_adata_from_indices, _get_final_indices


@nw.narwhalify
def _filter_df(df: IntoDataFrameT, *predicates: Predicates) -> IntoDataFrameT:
    return df.filter(*predicates)


def _filter_obs(adata: ad.AnnData, *predicates: Predicates) -> pd.Index:
    return _filter_df(adata.obs, *predicates).index


def _filter_var(adata: ad.AnnData, *predicates: Predicates) -> pd.Index:
    return _filter_df(adata.var, *predicates).index


def _filter_x(adata: ad.AnnData, *predicates: Predicates, layer: str | None = None) -> pd.Index:
    return _filter_df(adata.to_df(layer=layer), *predicates).index


def _filter_obs_names(adata: ad.AnnData, *predicates: Predicates) -> pd.Index:
    return _filter_df(adata.obs_names.to_frame(name="obs_names"), *predicates).index


def _filter_var_names(adata: ad.AnnData, *predicates: Predicates) -> pd.Index:
    return _filter_df(adata.var_names.to_frame(name="var_names"), *predicates).index


def _filter(
    adata: ad.AnnData,
    obs: Predicates | None = None,
    var: Predicates | None = None,
    x: Predicates | None = None,
    obs_names: Predicates | None = None,
    var_names: Predicates | None = None,
    layer: str | None = None,
    sparse: Literal["csr", "csc", True, False] | None = None,
) -> ad.AnnData:
    obs_names_idx = []
    var_names_idx = []

    if obs:
        obs_names_idx.append(_filter_obs(adata, obs))

    if obs_names:
        obs_names_idx.append(_filter_obs_names(adata, obs_names))

    if var:
        var_names_idx.append(_filter_var(adata, var))

    if var_names:
        var_names_idx.append(_filter_var_names(adata, var_names))

    if x:
        obs_names_idx.append(_filter_x(adata, x, layer=layer))

    final_obs_idx = adata.obs_names if not obs_names_idx else _get_final_indices(adata.obs_names, obs_names_idx)
    final_var_idx = adata.var_names if not var_names_idx else _get_final_indices(adata.var_names, var_names_idx)

    return _construct_adata_from_indices(adata, final_obs_idx, final_var_idx, sparse)



from collections.abc import Callable
from typing import Any, Literal, TypeAlias, TypeVar

import anndata as ad

from annsel.core.extensions import register_anndata_accessor
from annsel.core.typing import Predicates
from annsel.tl._filter import _filter
from annsel.tl._group_by import _group_by
from annsel.tl._select import _select

# Define type aliases
T = TypeVar("T")
GroupNames = tuple[str, ...]
GroupedResult: TypeAlias = ad.AnnData | tuple[GroupNames, ad.AnnData] | tuple[GroupNames, GroupNames, ad.AnnData]


@register_anndata_accessor("an")
class AnnselAccessor:
    """An extension for AnnData."""

    def __init__(self, adata: ad.AnnData):
        self._obj: ad.AnnData = adata

    def filter(
        self,
        obs: Predicates | None = None,
        var: Predicates | None = None,
        x: Predicates | None = None,
        obs_names: Predicates | None = None,
        var_names: Predicates | None = None,
        layer: str | None = None,
        sparse: Literal["csr", "csc", True, False] | None = None,
        copy: bool = False,
    ) -> ad.AnnData:
        """Filter the AnnData object by the given predicates.

        Parameters
        ----------
        obs
            Predicates to filter the observations by.
        var
            Predicates to filter the variables by.
        x
            Predicates to filter the data by.
        obs_names
            Predicates to filter the observation names by.
        var_names
            Predicates to filter the variable names by.
        layer
            The layer to filter.
        sparse
            Whether to use sparse matrices.
        copy
            Whether to return a copy of the AnnData object.

        Returns
        -------
        The filtered AnnData object.

        Examples
        --------
        >>> import annsel as an
        >>> adata = an.datasets.leukemic_bone_marrow_dataset()
        >>> adata.an.filter(
        ...     obs=(
        ...         an.col(["Cell_label"]).is_in(["Classical Monocytes", "CD8+CD103+ tissue resident memory T cells"]),
        ...         an.col(["sex"]) == "male",
        ...     ),
        ...     var=an.col(["vst.mean"]) >= 3,
        ... )
        AnnData object with n_obs × n_vars = 6741 × 458
             obs: 'Cluster_ID', 'donor_id', 'Sample_Tag', 'Cell_label', 'is_primary_data', 'organism_ontology_term_id', 'self_reported_ethnicity_ontology_term_id', 'assay_ontology_term_id', 'tissue_ontology_term_id', 'Genotype', 'development_stage_ontology_term_id', 'sex_ontology_term_id', 'disease_ontology_term_id', 'cell_type_ontology_term_id', 'suspension_type', 'tissue_type', 'cell_type', 'assay', 'disease', 'organism', 'sex', 'tissue', 'self_reported_ethnicity', 'development_stage', 'observation_joinid'
             var: 'vst.mean', 'vst.variance', 'vst.variance.expected', 'vst.variance.standardized', 'vst.variable', 'feature_is_filtered', 'Unnamed: 0', 'feature_name', 'feature_reference', 'feature_biotype', 'feature_length', 'feature_type'
             uns: 'cell_type_ontology_term_id_colors', 'citation', 'default_embedding', 'schema_reference', 'schema_version', 'title'
             obsm: 'X_bothumap', 'X_pca', 'X_projected', 'X_projectedmean', 'X_tsneni', 'X_umapni'

        """
        _adata = _filter(self._obj, obs, var, x, obs_names, var_names, layer, sparse)
        return _adata if copy else _adata.copy()

    def select(
        self,
        obs: Predicates | None = None,
        var: Predicates | None = None,
        x: Predicates | None = None,
        sparse: Literal["csr", "csc", True, False] | None = None,
        copy: bool = False,
    ) -> ad.AnnData:
        """Select the AnnData object by the given predicates.

        Parameters
        ----------
        obs
            Predicates to filter the observations by.
        var
            Predicates to filter the variables by.
        x
            Predicates to filter the data by.
        layer
            The layer to filter.
        sparse
            Whether to use sparse matrices.
        copy
            Whether to return a copy of the AnnData object.

        Returns
        -------
        The selected AnnData object.

        Examples
        --------
        >>> import annsel as an
        >>> adata = an.datasets.leukemic_bone_marrow_dataset()
        >>> adata.an.select(
        ...     obs=an.col(["Cell_label", "sex"]),
        ...     var=an.col(["feature_name"]),
        ...     x=an.col(["ENSG00000205336"]),
        ... )
        AnnData object with n_obs × n_vars = 31586 × 1
            obs: 'Cell_label', 'sex'
            var: 'feature_name'
            uns: 'cell_type_ontology_term_id_colors', 'citation', 'default_embedding', 'schema_reference', 'schema_version', 'title'
            obsm: 'X_bothumap', 'X_pca', 'X_projected', 'X_projectedmean', 'X_tsneni', 'X_umapni'
        """
        _adata = _select(self._obj, obs, var, x, sparse)
        return _adata if copy else _adata.copy()

    def group_by(
        self,
        obs: Predicates | None = None,
        var: Predicates | None = None,
        sparse: Literal["csr", "csc", True, False] | None = None,
        return_group_names: bool = False,
    ) -> GroupedResult:
        """Group the AnnData object by the given predicates.

        Parameters
        ----------
        obs
            Predicates to group the observations by.
        var
            Predicates to group the variables by.
        sparse
            Whether to use sparse matrices.
        return_group_names
            Whether to return the group names.

        Returns
        -------
        An iterator of grouped AnnData objects with the group names if `return_group_names` is `True`.

        Examples
        --------
        >>> import annsel as an
        >>> adata = an.datasets.leukemic_bone_marrow_dataset()
        >>> groups = adata.an.group_by(
        ...     obs=an.col(["Cell_label", "sex"]),
        ...     var=an.col(["feature_name"]),
        ...     return_group_names=False,
        ... )
        """
        gb_adata = _group_by(self._obj, obs, var, return_group_names=return_group_names, sparse_method=sparse)

        return gb_adata

    def pipe(self, func: Callable[..., T] | tuple[Callable[..., T], str], *args: Any, **kwargs: Any) -> Any:
        """
        Apply chainable functions ``func(self, *args, **kwargs)`` that expect :class:`~anndata.AnnData`.

        Parameters
        ----------
        func
            Function to apply to this :class:`~anndata.AnnData` object. ``args``, and ``kwargs`` are passed into ``func``.
            Alternatively a ``(callable, data_keyword)`` tuple where ``data_keyword`` is a string indicating
            the keyword of ``callable`` that expects the xarray object.
        *args
            Positional arguments passed into ``func``.
        **kwargs
            Keyword arguments passed into ``func``.

        Returns
        -------
            The return type of ``func``.

        Raises
        ------
        ValueError
            When the data target is both the pipe target and a keyword argument.

        Notes
        -----
        Use ``.pipe`` when chaining together functions that expect AnnData objects, e.g., instead of writing

        .. code:: python

            f(g(h(adata), arg1=a), arg2=b, arg3=cs)

        You can write

        .. code:: python

            (adata.an.pipe(h).pipe(g, arg1=a).an.pipe(f, arg2=b, arg3=c))

        If you have a function that takes the data as (say) the second argument, pass a tuple indicating
        which keyword expects the data. For example suppose ``f`` takes its data as ``arg2``

        .. code:: python

            (adata.an.pipe(h).pipe(g, arg1=a).an.pipe((f, "arg2"), arg1=a, arg3=c))

        See Also
        --------
        pandas.DataFrame.pipe
        xarray.DataArray.pipe
        xarray.Dataset.pipe

        Examples
        --------
        >>> import annsel as an
        >>> adata = an.datasets.leukemic_bone_marrow_dataset()
        >>> adata.an.pipe(sc.pl.embedding, basis="X_tsneni", color="Cell_label")
        """
        if isinstance(func, tuple):
            func, target = func
            if target in kwargs:
                raise ValueError(f"{target} is both the pipe target and a keyword argument")
            kwargs[target] = self._obj
        else:
            args = (self._obj, *args)
        return func(*args, **kwargs)
